#!/usr/bin/env python
#
# pKa calculations with APBS
#
# Copyright University College Dublin & Washington University St. Louis 2004-2007
# All rights reserved
#

#
# Get paths
#
debug=False
import optparse
import sys, os

from src.definitions import Definition
from src.forcefield import Forcefield
from src.routines import Routines
from src.protein import getPDBFile, readPDB, Protein, Amino, Nucleic
from src.aa import LIG
from src.errors import PDB2PKAError
from pdb2pka import inputgen_pKa

from pdb2pka.pka_routines import smooth
from pdb2pka import pka_help

def usage(x):
    #
    # Print usage guidelines
    #
    print 'Usage: pka.py --ff <forcefield> --lig <ligand in MOL2> --pdie <protein diel cons> --maps <1 for using provided 3D maps; 2 for genereting new maps>'
    print '--xdiel <xdiel maps> --ydiel <ydiel maps> --zdiel <zdiel maps> --kappa <ion-accessibility map> '
    print '--smooth <st.dev [A] of Gaussian smooting of 3D maps at the boundary, bandthwith=3 st.dev> <pdbfile>'
    print 'Force field can be amber, charmm and parse'
    print
    return

#
# --------------------------------------------------
#

def startpKa():
    """
        Function for starting pKa script from the command line.

        Returns
            protein:    The protein object as generated by PDB2PQR
            routines:   The routines object as generated by PDB2PQR
            forcefield: The forcefield object as generated by PDB2PQR
    """
    print
    print 'PDB2PQR pKa calculations'
    print

    parser = optparse.OptionParser()

    ##
    ## set optparse options
    ##
    parser.add_option(
        '-v','--verbose',
        dest='verbose',
        action="store_true",
        default=False,
        )
    parser.add_option(
        '--pdie',
        dest='pdie',
        default=8,
        type='int',
        help='<protein dielectric constant>',
        )
    parser.add_option(
        '--sdie',
        dest='sdie',
        default=80,
        type='int',
        help='<solvent dielectric constant>',
        )
    parser.add_option(
        '--ff',
        dest='ff',
        type='choice',
        default='parse',
        choices=("amber","AMBER","charmm","CHARMM","parse","PARSE",),
        help='<force field (amber, charmm, parse)>',
        )
    parser.add_option(
        '--resume',
        dest='resume',
        action="store_true",
        default=False,
        help='resume run from saved state.',
        )
    parser.add_option(
        '--ligand',
        dest='ligand',
        type='str',
        help='<ligand in MOL2 format>',
        )
    parser.add_option(
        '--maps',
        dest='maps',
        default=None,
        type='int',
        help='<1 for using provided 3D maps; 2 for genereting new maps>',
        )
    parser.add_option(
        '--xdiel',
        dest='xdiel',
        default=None,
        type='str',
        help='<xdiel maps>',
        )
    parser.add_option(
        '--ydiel',
        dest='ydiel',
        default=None,
        type='str',
        help='<ydiel maps>',
        )
    parser.add_option(
        '--zdiel',
        dest='zdiel',
        default=None,
        type='str',
        help='<zdiel maps>',
        )
    parser.add_option(
        '--kappa',
        dest='kappa',
        default=None,
        type='str',
        help='<ion-accessibility map>',
        )
    parser.add_option(
        '--smooth',
        dest='sd',
        default=None,
        type='float',
        help='<st.dev [A] of Gaussian smooting of 3D maps at the boundary, bandthwith=3 st.dev>',
        )
    #
    # Cut off energy for calculating non-charged-charged interaction energies
    #
    parser.add_option('--pairene',dest='pairene',type='float',default=1.0,
                      help='Cutoff energy in kT for calculating non charged-charged interaction energies. Default: %default')
    #
    # Options for doing partial calculations
    #
    parser.add_option('--res_energy',
                      dest='desolvation_res',
                      default=[],
                      action='append',
                      type='string',
                      help='Calculate desolvation energy and interaction energy for this residue in its default protonation state. Protonation states can be specified with the --protonation_state argument')
    parser.add_option('--PS_file',dest='PS_file',default='',type='string',action='store',help='Set protonation states according to the pdb2pka protonation state file (option --PS_file)')
    (options,args,) = parser.parse_args()

    ##
    ## parse optparse options
    ##
    ff = options.ff.lower()
    pdie = options.pdie
    verbose = options.verbose
    sdie = options.sdie
    maps = options.maps
    xdiel = options.xdiel
    ydiel = options.ydiel
    zdiel = options.zdiel
    kappa = options.kappa
    sd = options.sd

    #
    # Find the PDB file
    #
    if len(args) != 2:
        parser.error("Usage: pka.py [options] <pdbfile> <output directory>\n")
    input_path = args[0]
    output_path = args[1]

    ligand = None
    if options.ligand is not None:
        try:
            ligand = open(options.ligand, 'rU')
        except IOError:
            print 'Unable to find ligand file %s! Skipping...' % options.ligand

    #Set up the protien object
    #In the standalone version of pdb2pka this is redundent but needed so we emulate the
    #interface needed by pdb2pqr

    pdbfile = getPDBFile(input_path)
    pdblist, errlist = readPDB(pdbfile)
    if len(errlist) != 0 and verbose:
        print "Warning: %s is a non-standard PDB file.\n" %input_path
        print errlist
    #
    # Read the definition file
    #
    myDefinition = Definition()
    #
    #
    # Choose whether to include the ligand or not
    #
    # Add the ligand to the pdb2pqr arrays
    #
    if ligand is None:
        myProtein = Protein(pdblist, myDefinition)
    else:
        from pdb2pka.ligandclean import ligff
        myProtein, _, _ = ligff.initialize(myDefinition, ligand, pdblist, verbose)

    #
    # Call the pre_init function
    #
    return pre_init(protein=myProtein,
                    output_dir=output_path,
                    ff=ff,
                    verbose=verbose,
                    pdie=pdie,
                    sdie=sdie,
                    maps=maps,
                    xdiel=xdiel,
                    ydiel=ydiel,
                    zdiel=zdiel,
                    kappa=kappa,
                    sd=sd,
                    ligand=ligand),options


def pre_init(original_pdb_list=None,
             output_dir=None,
             ff=None,
             verbose=False,
             pdie=8.0,
             sdie=80,
             maps=None,
             xdiel=None,
             ydiel=None,
             zdiel=None,
             kappa=None,
             sd=None,
             ligand=None):
    """This function cleans the PDB and prepares the APBS input file

    Prepares the output folder."""

    #prepare the output directory

    output_dir = os.path.abspath(output_dir)

    try:
        os.makedirs(output_dir)
    except OSError:
        if not os.path.isdir(output_dir):
            raise ValueError('Target directory is a file! Aborting.')

    workspace_dir = os.path.join(output_dir,'workspace')

    try:
        os.makedirs(workspace_dir)
    except OSError:
        if not os.path.isdir(output_dir):
            raise ValueError('Target directory is a file! Aborting.')

    #
    # remove hydrogen atoms
    #

    working_pdb_filename = os.path.join(workspace_dir,'working.pdb')

    pka_help.dump_protein_no_hydrogens(original_pdb_list, working_pdb_filename)
    #
    # Get the PDBfile
    #
    pdbfile = getPDBFile(working_pdb_filename)
    pdblist, errlist = readPDB(pdbfile)

    if verbose:
        print "Beginning PDB2PKA...\n"
    #
    # Read the definition file
    #
    myDefinition = Definition()
    ligand_titratable_groups=None
    #
    #
    # Choose whether to include the ligand or not
    #
    # Add the ligand to the pdb2pqr arrays
    #
    Lig=None
    if ligand is None:
        myProtein = Protein(pdblist, myDefinition)
    else:
        from pdb2pka.ligandclean import ligff
        myProtein, myDefinition, Lig = ligff.initialize(myDefinition, ligand, pdblist, verbose)
    #
    # =======================================================================
    #
    # We have identified the structural elements, now contiue with the setup
    #
    # Print something for some reason?
    #
    if verbose:
        print "Created protein object -"
        print "\tNumber of residues in protein: %s" % myProtein.numResidues()
        print "\tNumber of atoms in protein   : %s" % myProtein.numAtoms()
    #
    # Set up all other routines
    #
    myRoutines = Routines(myProtein, verbose) #myDefinition)
    myRoutines.updateResidueTypes()
    myRoutines.updateSSbridges()
    myRoutines.updateBonds()
    myRoutines.setTermini()
    myRoutines.updateInternalBonds()

    myRoutines.applyNameScheme(Forcefield(ff, myDefinition, None))
    myRoutines.findMissingHeavy()
    myRoutines.addHydrogens()
    myRoutines.debumpProtein()

    #myRoutines.randomizeWaters()
    myProtein.reSerialize()
    #
    # Inject the information on hydrogen conformations in the HYDROGENS.DAT arrays
    # We get this information from ligand_titratable_groups
    #
    from src.hydrogens import hydrogenRoutines
    myRoutines.updateInternalBonds()
    myRoutines.calculateDihedralAngles()
    myhydRoutines = hydrogenRoutines(myRoutines)
    #
    # Here we should inject the info!!
    #
    myhydRoutines.setOptimizeableHydrogens()
    myhydRoutines.initializeFullOptimization()
    myhydRoutines.optimizeHydrogens()
    myhydRoutines.cleanup()
    myRoutines.setStates()

    #
    # Choose the correct forcefield
    #
    myForcefield = Forcefield(ff, myDefinition, None)
    if Lig:
        hitlist, misslist = myRoutines.applyForcefield(myForcefield)
        #
        # Can we get charges for the ligand?
        #
        templist=[]
        ligsuccess=False
        for residue in myProtein.getResidues():
            if isinstance(residue, LIG):
                templist = []
                Lig.make_up2date(residue)
                net_charge=0.0
                print 'Ligand',residue
                print 'Atom\tCharge\tRadius'
                for atom in residue.getAtoms():
                    if atom.mol2charge:
                        atom.ffcharge=atom.mol2charge
                    else:
                        atom.ffcharge = Lig.ligand_props[atom.name]["charge"]
                    #
                    # Find the net charge
                    #
                    net_charge=net_charge+atom.ffcharge
                    #
                    # Assign radius
                    #
                    atom.radius = Lig.ligand_props[atom.name]["radius"]
                    print '%s\t%6.4f\t%6.4f' %(atom.name,atom.ffcharge,atom.radius)
                    if atom in misslist:
                        misslist.pop(misslist.index(atom))
                        templist.append(atom)
                    #
                    # Store the charge and radius in the atom instance for later use
                    # This really should be done in a nicer way, but this will do for now
                    #
                    atom.secret_radius=atom.radius
                    atom.secret_charge=atom.ffcharge
                    #
                    #

                charge = residue.getCharge()
                if abs(charge - round(charge)) > 0.01:
                    # Ligand parameterization failed
                    myProtein.residues.remove(residue)
                    raise Exception('Non-integer charge on ligand: %8.5f' %charge)
                else:
                    ligsuccess = 1
                    # Mark these atoms as hits
                    hitlist = hitlist + templist
                #
                # Print the net charge
                #
                print 'Net charge for ligand %s is: %5.3f' %(residue.name,net_charge)
        #
        # Temporary fix; if ligand was successful, pull all ligands from misslist
        # Not sure if this is needed at all here ...? (Jens wrote this)
        #
        if ligsuccess:
            templist = misslist[:]
            for atom in templist:
                if isinstance(atom.residue, Amino) or isinstance(atom.residue, Nucleic):
                    continue
                misslist.remove(atom)

    if verbose:
        print "Created protein object (after processing myRoutines) -"
        print "\tNumber of residues in protein: %s" % myProtein.numResidues()
        print "\tNumber of atoms in protein   : %s" % myProtein.numAtoms()
    #
    # Create the APBS input file
    #
    import src.psize
    size=src.psize.Psize()

    method=""
    async=0
    split=0

    igen = inputgen_pKa.inputGen(working_pdb_filename)
    #
    # For convenience
    #
    igen.pdie = pdie
    print 'Setting protein dielectric constant to ',igen.pdie
    igen.sdie=sdie
    igen.maps=maps
    if maps==1:
        print "Using dielectric and mobile ion-accessibility function maps in PBE"
        if xdiel:
            igen.xdiel = xdiel
        else:
            raise PDB2PKAError('X dielectric map is missing')
        if ydiel:
            igen.ydiel = ydiel
        else:
            raise PDB2PKAError("Y dielectric map is missing\n")
        if zdiel:
            igen.zdiel = zdiel
        else:
            raise PDB2PKAError("Z dielectric map is missing\n")

        print 'Setting dielectric function maps: %s, %s, %s'%(igen.xdiel,igen.ydiel,igen.zdiel)

        if kappa:
            igen.kappa = kappa
        else:
            raise PDB2PKAError("Mobile ion-accessibility map is missing\n")

        print 'Setting mobile ion-accessibility function map to: ',igen.kappa

        if sd:
            xdiel_smooth, ydiel_smooth, zdiel_smooth = smooth(xdiel,ydiel,zdiel)
            igen.xdiel = xdiel_smooth
            igen.ydiel = ydiel_smooth
            igen.zdiel = zdiel_smooth
    #
    # Return all we need
    #
    return output_dir, myProtein, myRoutines, myForcefield,igen, ligand_titratable_groups, maps, sd

#
# --------------
#

if __name__ == "__main__":

    (output_dir, protein, routines, forcefield,apbs_setup, ligand_titratable_groups, maps, sd), options = startpKa()
    from pdb2pka import pka_routines
    mypkaRoutines = pka_routines.pKaRoutines(protein, routines, forcefield, apbs_setup, output_dir, maps, sd,
                                             restart=not options.resume, pairene=options.pairene)

    print 'Doing full pKa calculation'
    mypkaRoutines.runpKa()




"""
    Routines for PDB2PQR

    This module contains the protein object used in PDB2PQR and associated
    methods
    
    ----------------------------
   
    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, 
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice, 
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, 
		  this list of conditions and the following disclaimer in the documentation 
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------

"""

__date__ = "13 May 2008"
__author__ = "Todd Dolinsky, Yong Huang"

from pdb import *
from structures import *
from aa import *
from na import *


class Protein:
    """
    Protein class

    The protein class represents the parsed PDB, and provides a
    hierarchy of information - each Protein contains a list of Chain
    objects as provided in the PDB file.  Each Chain then contains its
    associated list of Residue objects, and each Residue contains a list
    of Atom objects, completing the hierarchy.
    """

    def __init__(self, pdblist, definition):
        """
        Initialize using parsed PDB file

        Parameters
            pdblist: List of Classes of PDB lines as created
        """
        self.chainmap = {}
        self.chains = []
        self.residues = []
        self.referencemap = definition.map
        self.patchmap = definition.patches

        chainDict = {}
        previousAtom = None
        residue = []
        numModels = 0
        numChains = 1
        count = 0

        for record in pdblist:  # Find number of chains
            if isinstance(record, TER):
                numChains += 1

        for record in pdblist:
            if isinstance(record, ATOM) or isinstance(record, HETATM):

                if (
                    record.chainID == ""
                    and numChains > 1
                    and record.resName not in ["WAT", "HOH"]
                ):
                    # Assign a chain ID
                    record.chainID = string.ascii_uppercase[count]

                chainID = record.chainID
                resSeq = record.resSeq
                resName = record.resName
                iCode = record.iCode

                if previousAtom == None:
                    previousAtom = record

                if chainID not in chainDict:
                    myChain = Chain(chainID)
                    chainDict[chainID] = myChain

                if (
                    resSeq != previousAtom.resSeq
                    or iCode != previousAtom.iCode
                    or chainID != previousAtom.chainID
                ):
                    myResidue = self.createResidue(residue, previousAtom.resName)
                    chainDict[previousAtom.chainID].addResidue(myResidue)
                    residue = []

                residue.append(record)
                previousAtom = record

            elif isinstance(record, END):
                myResidue = self.createResidue(residue, previousAtom.resName)
                chainDict[previousAtom.chainID].addResidue(myResidue)
                residue = []

            elif isinstance(record, MODEL):
                numModels += 1
                if residue == []:
                    continue
                if numModels > 1:
                    myResidue = self.createResidue(residue, previousAtom.resName)
                    chainDict[previousAtom.chainID].addResidue(myResidue)
                    break

            elif isinstance(record, TER):
                count += 1

        if residue != [] and numModels <= 1:
            myResidue = self.createResidue(residue, previousAtom.resName)
            chainDict[previousAtom.chainID].addResidue(myResidue)

        # Keep a map for accessing chains via chainID

        self.chainmap = chainDict.copy()

        # Make a list for sequential ordering of chains

        if chainDict.has_key(""):
            chainDict["ZZ"] = chainDict[""]
            del chainDict[""]

        keys = chainDict.keys()
        keys.sort()

        for key in keys:
            self.chains.append(chainDict[key])

        for chain in self.chains:
            # if chain.numResidues() == 1:
            # We cannot support Amino Acid chains with only one residue-
            # It is unclear whether they are Nterm, Cterm, or both.
            # residue = chain.residues[0]
            # if isinstance(residue, Amino):
            # raise ValueError, "Unable to support amino acid chains of only one residue (%s)" % residue

            for residue in chain.getResidues():
                self.residues.append(residue)

    def createResidue(self, residue, resname):
        """
        Create a residue object.  If the resname is a known residue
        type, try to make that specific object, otherwise just make
        a standard residue object.

        Parameters
            residue:  A list of atoms (list)
            resname:  The name of the residue (string)

        Returns:
            residue:  The residue object (Residue)
        """
        try:
            refobj = self.referencemap[resname]
            if refobj.name != resname:  # Patched!
                obj = "%s(residue, refobj)" % refobj.name
                residue = eval(obj)
                residue.reference = refobj
            else:
                obj = "%s(residue, refobj)" % resname
                residue = eval(obj)
        except (KeyError, NameError):
            residue = Residue(residue)
        return residue

    def printAtoms(self, atomlist, chainflag=False, pdbfile=False):
        """
        Get the text for the entire protein
        Parameters
            atomlist:  The list of atoms to include (list)
            chainflag: Flag whether to print chainid or not -
                          Defaults to False
        Returns
            text:      The list of (stringed) atoms (list)
        """
        self.reSerialize()
        text = []
        currentchainID = None
        for atom in atomlist:
            # Print the "TER" records between chains
            if currentchainID == None:
                currentchainID = atom.chainID
            elif atom.chainID != currentchainID:
                currentchainID = atom.chainID
                text.append("TER\n")

            if not atom.resName:
                continue

            if pdbfile == True:
                text.append("%s\n" % atom.getPDBString())
            else:
                text.append("%s\n" % atom.getPQRString(chainflag=chainflag))
        text.append("TER\nEND")
        return text

    def createHTMLTypeMap(self, definition, outfilename):
        """
        Create an HTML typemap file at the desired location. If a
        type cannot be found for an atom a blank is listed.

        Parameters
            definition: The definition objects.
            outfilename:  The name of the file to write (string)
        """
        from forcefield import Forcefield
        from aconf import STYLESHEET

        # Cache the initial atom numbers
        numcache = {}
        for atom in self.getAtoms():
            numcache[atom] = atom.serial
        self.reSerialize()

        amberff = Forcefield("amber", definition, None)
        charmmff = Forcefield("charmm", definition, None)

        file = open(outfilename, "w")
        file.write("<HTML>\n")
        file.write("<HEAD>\n")
        file.write("<TITLE>PQR Typemap (beta)</TITLE>\n")
        file.write('<link rel="stylesheet" href="%s" type="text/css">\n' % STYLESHEET)
        file.write("</HEAD>\n")
        file.write("<BODY>\n")
        file.write(
            "<H3>This is a developmental page including the atom type for the atoms in the PQR file.</H3><P>\n"
        )
        file.write("<TABLE CELLSPACING=2 CELLPADDING=2 BORDER=1>\n")
        file.write(
            "<tr><th>Atom Number</th><th>Atom Name</th><th>Residue Name</th><th>Chain ID</th><th>AMBER Atom Type</th><th>CHARMM Atom Type</th></tr>\n"
        )

        for atom in self.getAtoms():
            if isinstance(atom.residue, (Amino, WAT, Nucleic)):
                resname = atom.residue.ffname
            else:
                resname = atom.residue.name

            ambergroup = amberff.getGroup(resname, atom.name)
            charmmgroup = charmmff.getGroup(resname, atom.name)

            file.write(
                "<tr><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>\n"
                % (
                    atom.serial,
                    atom.name,
                    resname,
                    atom.chainID,
                    ambergroup,
                    charmmgroup,
                )
            )

        file.write("</table>\n")
        file.write("</BODY></HTML>\n")
        file.close()

        # Return the original numbers back
        for atom in self.getAtoms():
            atom.serial = numcache[atom]

        del numcache
        del amberff
        del charmmff

    def reSerialize(self):
        """
        Generate new serial numbers for atoms in the protein
        """
        count = 1
        for atom in self.getAtoms():
            atom.set("serial", count)
            count += 1

    def getResidues(self):
        """
        Return the list of residues in the entire protein
        """
        return self.residues

    def numResidues(self):
        """
        Get the number of residues for the entire protein (including
        multiple chains)

        Returns
            count:  Number of residues in the protein (int)
        """
        return len(self.getResidues())

    def numAtoms(self):
        """
        Get the number of atoms for the entire protein(including
        multiple chains)
        """
        return len(self.getAtoms())

    def getAtoms(self):
        """
        Return all Atom objects in list format

        Returns
            atomlist:  List of Atom objects in the protein (list)
        """

        atomlist = []
        for chain in self.chains:
            for atom in chain.getAtoms():
                atomlist.append(atom)
        return atomlist

    def getCharge(self):
        """
        Get the total charge on the protein
        NOTE:  Since the misslist is used to identify incorrect
               charge assignments, this routine does not list the
               3 and 5 termini of nucleic acid chains as having
               non-integer charge even though they are (correctly)
               non-integer.
        Returns:
            misslist: List of residues with non-integer
                      charges (list)
            charge:   The total charge on the protein (float)
        """
        charge = 0.0
        misslist = []
        for chain in self.chains:
            for residue in chain.get("residues"):
                rescharge = residue.getCharge()
                charge += rescharge
                if isinstance(residue, Nucleic):
                    if residue.is3term or residue.is5term:
                        continue
                if float("%i" % rescharge) != rescharge:
                    misslist.append(residue)
        return misslist, charge

    def getChains(self):
        """
        Get the chains object

        Returns
            chains: The list of chains in the protein (chain)
        """
        return self.chains

    def getSummary(self):
        output = []
        for chain in self.chains:
            output.append(chain.getSummary())
        return " ".join(output)



"""
    Contact extension

    Find all hydrogen bonds as determined by the DISTANCE cutoff below.
    Uses PDB2PQR to determine donors and acceptors, and displays
    all available bonds to stdout in a WHATIF-like format.

    Author:  Julie C. Mitchell
"""

__date__ = "April 2007"
__author__ = "Julie C. Mitchell"

from src.utilities import distance
from src.routines import Cells

DIST_CUTOFF = 3.5  # max distance


def usage():
    return "Print a list of contacts to {output-path}.con\n"


def run_extension(routines, outroot, options):
    """
    Print a list of contacts.

    Parameters
        routines:  A link to the routines object
        outroot:   The root of the output name
        options:   options object
    """
    outname = outroot + ".con"
    outfile = open(outname, "w")

    # Initialize - set nearby cells, donors/acceptors

    cellsize = int(DIST_CUTOFF + 1.0 + 1.0)
    protein = routines.protein
    routines.setDonorsAndAcceptors()
    routines.cells = Cells(cellsize)
    routines.cells.assignCells(protein)

    for thisatom in protein.getAtoms():
        # Grab the list of thisatoms
        if not thisatom.hdonor:
            continue
        thisatomhs = []
        for bond in thisatom.bonds:
            if bond.isHydrogen():
                thisatomhs.append(bond)
        if thisatomhs == []:
            continue

        # For each thisatom, grab all thatatomeptors

        count = 0
        closeatoms = routines.cells.getNearCells(thisatom)
        for thatatom in closeatoms:
            if thisatom.residue == thatatom.residue:
                continue  # comment this out to include interresidue contacts
            if thatatom.isHydrogen():
                continue
            thisdist = distance(thisatom.getCoords(), thatatom.getCoords())
            if thisdist <= DIST_CUTOFF:
                count = count + 1
                thisBstring = "S"
                thatBstring = "S"
                hscore = 0.0
                if thisatom.hdonor & thatatom.hacceptor:
                    hscore = 1.0
                if thisatom.hacceptor & thatatom.hdonor:
                    hscore = 1.0
                if thisatom.isBackbone():
                    thisBstring = "B"
                if thatatom.isBackbone():
                    thatBstring = "B"
                outfile.write(
                    "%4d %4d %-4s (%4d  ) %s     %-4s<>%4d %-4s (%4d  ) %s     %-4s D=%6.2f  H-ene=%6.2f  Sym=  (%s-%s)\n"
                    % (
                        count,
                        thisatom.residue.resSeq,
                        thisatom.residue.name,
                        thisatom.residue.resSeq,
                        thisatom.residue.chainID,
                        thisatom.name,
                        thatatom.residue.resSeq,
                        thatatom.residue.name,
                        thatatom.residue.resSeq,
                        thatatom.residue.chainID,
                        thatatom.name,
                        thisdist,
                        hscore,
                        thisBstring,
                        thatBstring,
                    )
                )

    routines.write("\n")
    outfile.close()



##################################################
# file: AppService_types.py
#
# schema types generated by "ZSI.generate.wsdl2python.WriteServiceModule"
#    ../ZSI-2.1-a1/scripts/wsdl2py wsdl/opal.wsdl
#
##################################################

import ZSI
import ZSI.TCcompound
from ZSI.schema import (
    LocalElementDeclaration,
    ElementDeclaration,
    TypeDefinition,
    GTD,
    GED,
)

##############################
# targetNamespace
# http://nbcr.sdsc.edu/opal/types
##############################


class ns0:
    targetNamespace = "http://nbcr.sdsc.edu/opal/types"

    class ParamType_Def(ZSI.TC.String, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "ParamType")

        def __init__(self, pname, **kw):
            ZSI.TC.String.__init__(self, pname, pyclass=None, **kw)

            class Holder(str):
                typecode = self

            self.pyclass = Holder

    class IOType_Def(ZSI.TC.String, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "IOType")

        def __init__(self, pname, **kw):
            ZSI.TC.String.__init__(self, pname, pyclass=None, **kw)

            class Holder(str):
                typecode = self

            self.pyclass = Holder

    class FlagsType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "FlagsType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.FlagsType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="id",
                    aname="_id",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="tag",
                    aname="_tag",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.Boolean(
                    pname="default",
                    aname="_default",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="textDesc",
                    aname="_textDesc",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._id = None
                    self._tag = None
                    self._default = None
                    self._textDesc = None
                    return

            Holder.__name__ = "FlagsType_Holder"
            self.pyclass = Holder

    class ParamsType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "ParamsType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.ParamsType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="id",
                    aname="_id",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="tag",
                    aname="_tag",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="default",
                    aname="_default",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "ParamType", lazy=False)(
                    pname="paramType",
                    aname="_paramType",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "IOType", lazy=False)(
                    pname="ioType",
                    aname="_ioType",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.Boolean(
                    pname="required",
                    aname="_required",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="value",
                    aname="_value",
                    minOccurs=0,
                    maxOccurs="unbounded",
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="semanticType",
                    aname="_semanticType",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="textDesc",
                    aname="_textDesc",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._id = None
                    self._tag = None
                    self._default = None
                    self._paramType = None
                    self._ioType = None
                    self._required = None
                    self._value = []
                    self._semanticType = None
                    self._textDesc = None
                    return

            Holder.__name__ = "ParamsType_Holder"
            self.pyclass = Holder

    class GroupsType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "GroupsType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.GroupsType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="name",
                    aname="_name",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="elements",
                    aname="_elements",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.Boolean(
                    pname="required",
                    aname="_required",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.Boolean(
                    pname="exclusive",
                    aname="_exclusive",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="semanticType",
                    aname="_semanticType",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="textDesc",
                    aname="_textDesc",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._name = None
                    self._elements = None
                    self._required = None
                    self._exclusive = None
                    self._semanticType = None
                    self._textDesc = None
                    return

            Holder.__name__ = "GroupsType_Holder"
            self.pyclass = Holder

    class ImplicitParamsType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "ImplicitParamsType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.ImplicitParamsType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="id",
                    aname="_id",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="name",
                    aname="_name",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="extension",
                    aname="_extension",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "IOType", lazy=False)(
                    pname="ioType",
                    aname="_ioType",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.Boolean(
                    pname="required",
                    aname="_required",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="semanticType",
                    aname="_semanticType",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="textDesc",
                    aname="_textDesc",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TCnumbers.Iint(
                    pname="min",
                    aname="_min",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TCnumbers.Iint(
                    pname="max",
                    aname="_max",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._id = None
                    self._name = None
                    self._extension = None
                    self._ioType = None
                    self._required = None
                    self._semanticType = None
                    self._textDesc = None
                    self._min = None
                    self._max = None
                    return

            Holder.__name__ = "ImplicitParamsType_Holder"
            self.pyclass = Holder

    class FlagsArrayType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "FlagsArrayType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.FlagsArrayType_Def.schema
            TClist = [
                GTD("http://nbcr.sdsc.edu/opal/types", "FlagsType", lazy=False)(
                    pname="flag",
                    aname="_flag",
                    minOccurs=0,
                    maxOccurs="unbounded",
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                )
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._flag = []
                    return

            Holder.__name__ = "FlagsArrayType_Holder"
            self.pyclass = Holder

    class ParamsArrayType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "ParamsArrayType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.ParamsArrayType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="separator",
                    aname="_separator",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "ParamsType", lazy=False)(
                    pname="param",
                    aname="_param",
                    minOccurs=0,
                    maxOccurs="unbounded",
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._separator = None
                    self._param = []
                    return

            Holder.__name__ = "ParamsArrayType_Holder"
            self.pyclass = Holder

    class GroupsArrayType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "GroupsArrayType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.GroupsArrayType_Def.schema
            TClist = [
                GTD("http://nbcr.sdsc.edu/opal/types", "GroupsType", lazy=False)(
                    pname="group",
                    aname="_group",
                    minOccurs=0,
                    maxOccurs="unbounded",
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                )
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._group = []
                    return

            Holder.__name__ = "GroupsArrayType_Holder"
            self.pyclass = Holder

    class ImplicitParamsArrayType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "ImplicitParamsArrayType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.ImplicitParamsArrayType_Def.schema
            TClist = [
                GTD(
                    "http://nbcr.sdsc.edu/opal/types", "ImplicitParamsType", lazy=False
                )(
                    pname="param",
                    aname="_param",
                    minOccurs=0,
                    maxOccurs="unbounded",
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                )
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._param = []
                    return

            Holder.__name__ = "ImplicitParamsArrayType_Holder"
            self.pyclass = Holder

    class ArgumentsType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "ArgumentsType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.ArgumentsType_Def.schema
            TClist = [
                GTD("http://nbcr.sdsc.edu/opal/types", "FlagsArrayType", lazy=False)(
                    pname="flags",
                    aname="_flags",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "ParamsArrayType", lazy=False)(
                    pname="taggedParams",
                    aname="_taggedParams",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "ParamsArrayType", lazy=False)(
                    pname="untaggedParams",
                    aname="_untaggedParams",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD(
                    "http://nbcr.sdsc.edu/opal/types",
                    "ImplicitParamsArrayType",
                    lazy=False,
                )(
                    pname="implicitParams",
                    aname="_implicitParams",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "GroupsArrayType", lazy=False)(
                    pname="groups",
                    aname="_groups",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._flags = None
                    self._taggedParams = None
                    self._untaggedParams = None
                    self._implicitParams = None
                    self._groups = None
                    return

            Holder.__name__ = "ArgumentsType_Holder"
            self.pyclass = Holder

    class AppMetadataInputType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "AppMetadataInputType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.AppMetadataInputType_Def.schema
            TClist = []
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    return

            Holder.__name__ = "AppMetadataInputType_Holder"
            self.pyclass = Holder

    class AppMetadataType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "AppMetadataType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.AppMetadataType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="usage",
                    aname="_usage",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="info",
                    aname="_info",
                    minOccurs=0,
                    maxOccurs="unbounded",
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "ArgumentsType", lazy=False)(
                    pname="types",
                    aname="_types",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            else:
                # attribute handling code
                self.attribute_typecode_dict["appName"] = ZSI.TC.String()
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._usage = None
                    self._info = []
                    self._types = None
                    return

            Holder.__name__ = "AppMetadataType_Holder"
            self.pyclass = Holder

    class AppConfigInputType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "AppConfigInputType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.AppConfigInputType_Def.schema
            TClist = []
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    return

            Holder.__name__ = "AppConfigInputType_Holder"
            self.pyclass = Holder

    class AppConfigType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "AppConfigType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.AppConfigType_Def.schema
            TClist = [
                GTD("http://nbcr.sdsc.edu/opal/types", "AppMetadataType", lazy=False)(
                    pname="metadata",
                    aname="_metadata",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="binaryLocation",
                    aname="_binaryLocation",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="defaultArgs",
                    aname="_defaultArgs",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.Boolean(
                    pname="parallel",
                    aname="_parallel",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._metadata = None
                    self._binaryLocation = None
                    self._defaultArgs = None
                    self._parallel = None
                    return

            Holder.__name__ = "AppConfigType_Holder"
            self.pyclass = Holder

    class InputFileType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "InputFileType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.InputFileType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="name",
                    aname="_name",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.Base64String(
                    pname="contents",
                    aname="_contents",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._name = None
                    self._contents = None
                    return

            Holder.__name__ = "InputFileType_Holder"
            self.pyclass = Holder

    class JobInputType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "JobInputType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.JobInputType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="argList",
                    aname="_argList",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TCnumbers.Iint(
                    pname="numProcs",
                    aname="_numProcs",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "InputFileType", lazy=False)(
                    pname="inputFile",
                    aname="_inputFile",
                    minOccurs=0,
                    maxOccurs="unbounded",
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._argList = None
                    self._numProcs = None
                    self._inputFile = []
                    return

            Holder.__name__ = "JobInputType_Holder"
            self.pyclass = Holder

    class StatusOutputType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "StatusOutputType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.StatusOutputType_Def.schema
            TClist = [
                ZSI.TCnumbers.Iint(
                    pname="code",
                    aname="_code",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="message",
                    aname="_message",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.URI(
                    pname="baseURL",
                    aname="_baseURL",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._code = None
                    self._message = None
                    self._baseURL = None
                    return

            Holder.__name__ = "StatusOutputType_Holder"
            self.pyclass = Holder

    class JobSubOutputType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "JobSubOutputType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.JobSubOutputType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="jobID",
                    aname="_jobID",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "StatusOutputType", lazy=False)(
                    pname="status",
                    aname="_status",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._jobID = None
                    self._status = None
                    return

            Holder.__name__ = "JobSubOutputType_Holder"
            self.pyclass = Holder

    class OutputsByNameInputType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "OutputsByNameInputType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.OutputsByNameInputType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="jobID",
                    aname="_jobID",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.String(
                    pname="fileName",
                    aname="_fileName",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._jobID = None
                    self._fileName = None
                    return

            Holder.__name__ = "OutputsByNameInputType_Holder"
            self.pyclass = Holder

    class OutputFileType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "OutputFileType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.OutputFileType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="name",
                    aname="_name",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.URI(
                    pname="url",
                    aname="_url",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._name = None
                    self._url = None
                    return

            Holder.__name__ = "OutputFileType_Holder"
            self.pyclass = Holder

    class JobOutputType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "JobOutputType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.JobOutputType_Def.schema
            TClist = [
                ZSI.TC.URI(
                    pname="stdOut",
                    aname="_stdOut",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                ZSI.TC.URI(
                    pname="stdErr",
                    aname="_stdErr",
                    minOccurs=0,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "OutputFileType", lazy=False)(
                    pname="outputFile",
                    aname="_outputFile",
                    minOccurs=0,
                    maxOccurs="unbounded",
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._stdOut = None
                    self._stdErr = None
                    self._outputFile = []
                    return

            Holder.__name__ = "JobOutputType_Holder"
            self.pyclass = Holder

    class BlockingOutputType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "BlockingOutputType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.BlockingOutputType_Def.schema
            TClist = [
                GTD("http://nbcr.sdsc.edu/opal/types", "StatusOutputType", lazy=False)(
                    pname="status",
                    aname="_status",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
                GTD("http://nbcr.sdsc.edu/opal/types", "JobOutputType", lazy=False)(
                    pname="jobOut",
                    aname="_jobOut",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                ),
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._status = None
                    self._jobOut = None
                    return

            Holder.__name__ = "BlockingOutputType_Holder"
            self.pyclass = Holder

    class FaultType_Def(ZSI.TCcompound.ComplexType, TypeDefinition):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = (schema, "FaultType")

        def __init__(
            self, pname, ofwhat=(), attributes=None, extend=False, restrict=False, **kw
        ):
            ns = ns0.FaultType_Def.schema
            TClist = [
                ZSI.TC.String(
                    pname="message",
                    aname="_message",
                    minOccurs=1,
                    maxOccurs=1,
                    nillable=False,
                    typed=False,
                    encoded=kw.get("encoded"),
                )
            ]
            self.attribute_typecode_dict = attributes or {}
            if extend:
                TClist += ofwhat
            if restrict:
                TClist = ofwhat
            ZSI.TCcompound.ComplexType.__init__(
                self, None, TClist, pname=pname, inorder=0, **kw
            )

            class Holder:
                typecode = self

                def __init__(self):
                    # pyclass
                    self._message = None
                    return

            Holder.__name__ = "FaultType_Holder"
            self.pyclass = Holder

    class getAppMetadataInput_Dec(ElementDeclaration):
        literal = "getAppMetadataInput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "getAppMetadataInput")
            kw["aname"] = "_getAppMetadataInput"
            if (
                ns0.AppMetadataInputType_Def
                not in ns0.getAppMetadataInput_Dec.__bases__
            ):
                bases = list(ns0.getAppMetadataInput_Dec.__bases__)
                bases.insert(0, ns0.AppMetadataInputType_Def)
                ns0.getAppMetadataInput_Dec.__bases__ = tuple(bases)

            ns0.AppMetadataInputType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "getAppMetadataInput_Dec_Holder"

    class getAppMetadataOutput_Dec(ElementDeclaration):
        literal = "getAppMetadataOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "getAppMetadataOutput")
            kw["aname"] = "_getAppMetadataOutput"
            if ns0.AppMetadataType_Def not in ns0.getAppMetadataOutput_Dec.__bases__:
                bases = list(ns0.getAppMetadataOutput_Dec.__bases__)
                bases.insert(0, ns0.AppMetadataType_Def)
                ns0.getAppMetadataOutput_Dec.__bases__ = tuple(bases)

            ns0.AppMetadataType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "getAppMetadataOutput_Dec_Holder"

    class getAppConfigInput_Dec(ElementDeclaration):
        literal = "getAppConfigInput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "getAppConfigInput")
            kw["aname"] = "_getAppConfigInput"
            if ns0.AppConfigInputType_Def not in ns0.getAppConfigInput_Dec.__bases__:
                bases = list(ns0.getAppConfigInput_Dec.__bases__)
                bases.insert(0, ns0.AppConfigInputType_Def)
                ns0.getAppConfigInput_Dec.__bases__ = tuple(bases)

            ns0.AppConfigInputType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "getAppConfigInput_Dec_Holder"

    class getAppConfigOutput_Dec(ElementDeclaration):
        literal = "getAppConfigOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "getAppConfigOutput")
            kw["aname"] = "_getAppConfigOutput"
            if ns0.AppConfigType_Def not in ns0.getAppConfigOutput_Dec.__bases__:
                bases = list(ns0.getAppConfigOutput_Dec.__bases__)
                bases.insert(0, ns0.AppConfigType_Def)
                ns0.getAppConfigOutput_Dec.__bases__ = tuple(bases)

            ns0.AppConfigType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "getAppConfigOutput_Dec_Holder"

    class launchJobInput_Dec(ElementDeclaration):
        literal = "launchJobInput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "launchJobInput")
            kw["aname"] = "_launchJobInput"
            if ns0.JobInputType_Def not in ns0.launchJobInput_Dec.__bases__:
                bases = list(ns0.launchJobInput_Dec.__bases__)
                bases.insert(0, ns0.JobInputType_Def)
                ns0.launchJobInput_Dec.__bases__ = tuple(bases)

            ns0.JobInputType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "launchJobInput_Dec_Holder"

    class launchJobOutput_Dec(ElementDeclaration):
        literal = "launchJobOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "launchJobOutput")
            kw["aname"] = "_launchJobOutput"
            if ns0.JobSubOutputType_Def not in ns0.launchJobOutput_Dec.__bases__:
                bases = list(ns0.launchJobOutput_Dec.__bases__)
                bases.insert(0, ns0.JobSubOutputType_Def)
                ns0.launchJobOutput_Dec.__bases__ = tuple(bases)

            ns0.JobSubOutputType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "launchJobOutput_Dec_Holder"

    class launchJobBlockingInput_Dec(ElementDeclaration):
        literal = "launchJobBlockingInput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "launchJobBlockingInput")
            kw["aname"] = "_launchJobBlockingInput"
            if ns0.JobInputType_Def not in ns0.launchJobBlockingInput_Dec.__bases__:
                bases = list(ns0.launchJobBlockingInput_Dec.__bases__)
                bases.insert(0, ns0.JobInputType_Def)
                ns0.launchJobBlockingInput_Dec.__bases__ = tuple(bases)

            ns0.JobInputType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "launchJobBlockingInput_Dec_Holder"

    class launchJobBlockingOutput_Dec(ElementDeclaration):
        literal = "launchJobBlockingOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "launchJobBlockingOutput")
            kw["aname"] = "_launchJobBlockingOutput"
            if (
                ns0.BlockingOutputType_Def
                not in ns0.launchJobBlockingOutput_Dec.__bases__
            ):
                bases = list(ns0.launchJobBlockingOutput_Dec.__bases__)
                bases.insert(0, ns0.BlockingOutputType_Def)
                ns0.launchJobBlockingOutput_Dec.__bases__ = tuple(bases)

            ns0.BlockingOutputType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "launchJobBlockingOutput_Dec_Holder"

    class queryStatusInput_Dec(ZSI.TC.String, ElementDeclaration):
        literal = "queryStatusInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "queryStatusInput")
            kw["aname"] = "_queryStatusInput"

            class IHolder(str):
                typecode = self

            kw["pyclass"] = IHolder
            IHolder.__name__ = "_queryStatusInput_immutable_holder"
            ZSI.TC.String.__init__(self, **kw)

    class queryStatusOutput_Dec(ElementDeclaration):
        literal = "queryStatusOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "queryStatusOutput")
            kw["aname"] = "_queryStatusOutput"
            if ns0.StatusOutputType_Def not in ns0.queryStatusOutput_Dec.__bases__:
                bases = list(ns0.queryStatusOutput_Dec.__bases__)
                bases.insert(0, ns0.StatusOutputType_Def)
                ns0.queryStatusOutput_Dec.__bases__ = tuple(bases)

            ns0.StatusOutputType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "queryStatusOutput_Dec_Holder"

    class getOutputsInput_Dec(ZSI.TC.String, ElementDeclaration):
        literal = "getOutputsInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "getOutputsInput")
            kw["aname"] = "_getOutputsInput"

            class IHolder(str):
                typecode = self

            kw["pyclass"] = IHolder
            IHolder.__name__ = "_getOutputsInput_immutable_holder"
            ZSI.TC.String.__init__(self, **kw)

    class getOutputsOutput_Dec(ElementDeclaration):
        literal = "getOutputsOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "getOutputsOutput")
            kw["aname"] = "_getOutputsOutput"
            if ns0.JobOutputType_Def not in ns0.getOutputsOutput_Dec.__bases__:
                bases = list(ns0.getOutputsOutput_Dec.__bases__)
                bases.insert(0, ns0.JobOutputType_Def)
                ns0.getOutputsOutput_Dec.__bases__ = tuple(bases)

            ns0.JobOutputType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "getOutputsOutput_Dec_Holder"

    class getOutputAsBase64ByNameInput_Dec(ElementDeclaration):
        literal = "getOutputAsBase64ByNameInput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = (
                "http://nbcr.sdsc.edu/opal/types",
                "getOutputAsBase64ByNameInput",
            )
            kw["aname"] = "_getOutputAsBase64ByNameInput"
            if (
                ns0.OutputsByNameInputType_Def
                not in ns0.getOutputAsBase64ByNameInput_Dec.__bases__
            ):
                bases = list(ns0.getOutputAsBase64ByNameInput_Dec.__bases__)
                bases.insert(0, ns0.OutputsByNameInputType_Def)
                ns0.getOutputAsBase64ByNameInput_Dec.__bases__ = tuple(bases)

            ns0.OutputsByNameInputType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "getOutputAsBase64ByNameInput_Dec_Holder"

    class getOutputAsBase64ByNameOutput_Dec(ZSI.TC.Base64String, ElementDeclaration):
        literal = "getOutputAsBase64ByNameOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, **kw):
            kw["pname"] = (
                "http://nbcr.sdsc.edu/opal/types",
                "getOutputAsBase64ByNameOutput",
            )
            kw["aname"] = "_getOutputAsBase64ByNameOutput"

            class IHolder(str):
                typecode = self

            kw["pyclass"] = IHolder
            IHolder.__name__ = "_getOutputAsBase64ByNameOutput_immutable_holder"
            ZSI.TC.Base64String.__init__(self, **kw)

    class destroyInput_Dec(ZSI.TC.String, ElementDeclaration):
        literal = "destroyInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "destroyInput")
            kw["aname"] = "_destroyInput"

            class IHolder(str):
                typecode = self

            kw["pyclass"] = IHolder
            IHolder.__name__ = "_destroyInput_immutable_holder"
            ZSI.TC.String.__init__(self, **kw)

    class destroyOutput_Dec(ElementDeclaration):
        literal = "destroyOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "destroyOutput")
            kw["aname"] = "_destroyOutput"
            if ns0.StatusOutputType_Def not in ns0.destroyOutput_Dec.__bases__:
                bases = list(ns0.destroyOutput_Dec.__bases__)
                bases.insert(0, ns0.StatusOutputType_Def)
                ns0.destroyOutput_Dec.__bases__ = tuple(bases)

            ns0.StatusOutputType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "destroyOutput_Dec_Holder"

    class opalFaultOutput_Dec(ElementDeclaration):
        literal = "opalFaultOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"
        substitutionGroup = None

        def __init__(self, **kw):
            kw["pname"] = ("http://nbcr.sdsc.edu/opal/types", "opalFaultOutput")
            kw["aname"] = "_opalFaultOutput"
            if ns0.FaultType_Def not in ns0.opalFaultOutput_Dec.__bases__:
                bases = list(ns0.opalFaultOutput_Dec.__bases__)
                bases.insert(0, ns0.FaultType_Def)
                ns0.opalFaultOutput_Dec.__bases__ = tuple(bases)

            ns0.FaultType_Def.__init__(self, **kw)
            if self.pyclass is not None:
                self.pyclass.__name__ = "opalFaultOutput_Dec_Holder"


# end class ns0 (tns: http://nbcr.sdsc.edu/opal/types)
